# mostly taken from viscid... thanks Kris
from __future__ import annotations

import logging
import os
from collections import OrderedDict
from collections.abc import Mapping
from typing import Any

import numpy as np
from numpy.typing import NDArray

from ggcmpy import _jrrle, openggcm  # type: ignore[attr-defined]

from .fortran_file import FortranFile

logger = logging.getLogger(__name__)

read_ascii = False

_jrrle_read_func = (
    _jrrle.read_jrrle1d,
    _jrrle.read_jrrle2d,
    _jrrle.read_jrrle3d,
)


def _jrrle_read_field(file, fld_name: str, meta: dict[str, Any]) -> NDArray[Any]:
    shape = meta["shape"]
    ndim = len(shape)
    arr = np.empty(shape, dtype="float32", order="F")
    success = _jrrle_read_func[ndim - 1](file.unit, arr, fld_name, read_ascii)
    if not success:
        msg = "read_func failed"
        raise RuntimeError(msg)
    return arr


def _jrrle_inquire_next(
    file: FortranFile,
) -> tuple[str, dict[str, Any]] | None:
    b_varname = np.array(" " * 80, dtype="S80")
    b_tstring = np.array(" " * 80, dtype="S80")
    found_field, ndim, nx, ny, nz, it = _jrrle.inquire_next(
        file.unit, b_varname, b_tstring
    )
    if not found_field:
        return None

    shape = (nx, ny, nz)[:ndim]
    varname = str(np.char.decode(b_varname)).strip()
    timestr = str(np.char.decode(b_tstring)).strip()
    parsed = openggcm.parse_timestring(timestr)

    return varname, {
        "shape": shape,
        "inttime": it,
        "time": parsed["time"],
        "elapsed_time": parsed["elapsed_time"],
        "file_position": file.tell(),
    }


class JrrleFile(FortranFile):
    """Interface for actually opening / reading a jrrle file"""

    def __init__(self, filename: str | os.PathLike[Any], mode: str = "r"):
        assert mode == "r"

        self.fields_seen: OrderedDict[str, Any] = OrderedDict()
        self.seen_all_fields = False
        super().__init__(filename)

    def read_field(self, fld_name) -> tuple[Any, NDArray[Any]]:
        """Read a field"""
        logger.debug("read_field(%s)" % fld_name)
        meta = self._inquire(fld_name)
        return meta, _jrrle_read_field(self, fld_name, meta)

    @property
    def vars(self) -> Mapping[str, Mapping[str, Any]]:
        self._inquire_all_fields()
        return self.fields_seen

    def _inquire_all_fields(self) -> None:
        if self.seen_all_fields:
            return

        logger.debug("_inquire_all_fields() (%s)" % self.filename)
        self.rewind()
        while True:
            rv = self._inquire_next()
            if not rv:
                break
            self.advance_one_line()

    def _inquire(self, fld_name: str) -> Any:
        if fld_name in self.fields_seen:
            meta = self.fields_seen[fld_name]
            # For some mysterious reason, the rewind is necessary -- without it,
            # reading after the seek fails. Mysteriously, a self.tell() also
            # sometimes makes things work.
            self.rewind()
            self.seek(meta["file_position"])
            return meta

        if self.fields_seen:
            # seek to past last previously read field
            last_added = next(reversed(self.fields_seen))
            self.seek(self.fields_seen[last_added]["file_position"])
            self.advance_one_line()

        while True:
            rv = self._inquire_next()
            if not rv:
                break
            found_fld_name, meta = rv
            if found_fld_name == fld_name:
                return meta
            self.advance_one_line()

        msg = f"file '{self.filename}' has no field '{fld_name}'"
        raise KeyError(msg)

    def _inquire_next(self) -> tuple[str, Any] | None:
        """Collect the meta-data from the next field in the file

        Returns:
            tuple (field name, dict of meta data) both
            of which will be None if there are no more Fields

        Note:
            After this operation is done, the file-pointer will be reset
            to the position it was before the inquiry.
        """

        rv = _jrrle_inquire_next(self)
        if not rv:
            # end of file
            self.seen_all_fields = True
            return None

        varname, meta = rv
        if varname in self.fields_seen:
            assert meta == self.fields_seen[varname]
        else:
            self.fields_seen[varname] = meta

        return varname, meta



from __future__ import annotations

import datetime as dt
import os
from collections.abc import Hashable, Iterable, Mapping, Sequence
from itertools import islice
from pathlib import Path
from typing import Any
from warnings import warn

import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import ArrayLike, DTypeLike, NDArray
from typing_extensions import override
from xarray.coding.times import CFDatetimeCoder


def read_grid2(filename: os.PathLike[Any] | str) -> dict[str, NDArray[Any]]:
    # load the cell centered grid
    with Path(filename).open(encoding="utf-8") as fin:
        nx = int(next(fin).split()[0])
        gx = list(islice(fin, 0, nx, 1))

        ny = int(next(fin).split()[0])
        gy = list(islice(fin, 0, ny, 1))

        nz = int(next(fin).split()[0])
        gz = list(islice(fin, 0, nz, 1))

    return {
        "x": np.array(gx, dtype="f4"),
        "y": np.array(gy, dtype="f4"),
        "z": np.array(gz, dtype="f4"),
    }


def parse_timestring(timestr: str) -> dict[str, Any]:
    if not timestr.startswith("time="):
        msg = f"Time string '{timestr}' is malformed"
        raise ValueError(msg)
    time_comps = timestr.removeprefix("time=").split()

    return {
        "elapsed_time": float(time_comps[0]),
        "time": np.datetime64(
            dt.datetime.strptime(time_comps[2], "%Y:%m:%d:%H:%M:%S.%f")
        ),
    }


def decode_openggcm(ds: xr.Dataset) -> xr.Dataset:
    warn(
        "decode_openggcm() is deprecated. Pass 'decode_times=openggcm.AmieTimeArrayDecoder()' "
        "to open_dataset() instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    coder = AmieTimeArrayCoder()
    for name, var in ds.variables.items():
        ds[name] = coder.decode(var, name)

    return ds


def encode_openggcm(
    variables: Mapping[str, xr.Variable], attributes: Mapping[str, Any]
) -> tuple[Mapping[str, xr.Variable], Mapping[str, Any]]:
    coder = AmieTimeArrayCoder()
    new_variables = {name: coder.encode(var, name) for name, var in variables.items()}

    return new_variables, attributes


class AmieTimeArrayCoder(CFDatetimeCoder):
    """
    A custom coder for encoding and decoding time arrays in xarray Variables.
    This class extends the CFDatetimeCoder to handle variables with a custom
    "time_array" encoding, and otherwise falls back to the usual CF coding/encoding.

    Methods
    -------
    encode(variable: xr.Variable, name: Hashable | None = None) -> xr.Variable
        Encodes an xarray Variable with a "time_array" unit into a custom format.
    decode(variable: xr.Variable, name: Hashable | None = None) -> xr.Variable
        Decodes an xarray Variable with a "time_array" unit from a custom format.
    """

    def __init__(self, use_cftime: bool | None = None):
        super().__init__(use_cftime=use_cftime)

    @override
    def encode(
        self, variable: xr.Variable, name: Hashable | None = None
    ) -> xr.Variable:
        if variable.encoding.get("units") == "time_array":
            attrs = variable.attrs.copy()
            attrs["units"] = "time_array"
            attrs.pop("_FillValue", None)

            return xr.Variable(
                dims=(*variable.dims, "time_array"),
                data=_dt64_to_time_array(
                    variable,
                    variable.encoding.get("dtype", "int32"),
                ),
                attrs=attrs,
            )

        return super().encode(variable, name)

    @override
    def decode(
        self, variable: xr.Variable, name: Hashable | None = None
    ) -> xr.Variable:
        if variable.attrs.get("units") == "time_array":
            times: Any = variable.to_numpy().tolist()
            if variable.ndim == 1:
                times = _time_array_to_dt64([times])[0]
            else:
                times = _time_array_to_dt64(times)

            dims = (dim for dim in variable.dims if dim != "time_array")
            attrs = variable.attrs.copy()
            encoding = {"units": attrs.pop("units"), "dtype": variable.dtype}
            return xr.Variable(dims=dims, data=times, attrs=attrs, encoding=encoding)

        return super().decode(variable, name)


def _time_array_to_dt64(times: Iterable[Sequence[int]]) -> Sequence[np.datetime64]:
    return [
        np.datetime64(
            dt.datetime(
                year=time[0],
                month=time[1],
                day=time[2],
                hour=time[3],
                minute=time[4],
                second=time[5],
                microsecond=time[6] * 1000,
            ),
            "ns",
        )
        for time in times
    ]


def _dt64_to_time_array(times: ArrayLike, dtype: DTypeLike) -> ArrayLike:
    dt_times = pd.to_datetime(np.asarray(times))
    return np.array(
        [
            dt_times.year,
            dt_times.month,
            dt_times.day,
            dt_times.hour,
            dt_times.minute,
            dt_times.second,
            dt_times.microsecond // 1000,
        ],
        dtype=dtype,
    ).T


@xr.register_dataarray_accessor("ggcm")  # type: ignore[no-untyped-call]
@xr.register_dataset_accessor("ggcm")  # type: ignore[no-untyped-call]
class OpenGGCMAccessor:
    """
    Xarray accessor to add OpenGGCM-specific features

    As of now, that is ds.ggcm.coords which provides mlts and colats.
    """

    def __init__(self, xarray_obj: Any):
        self._obj = xarray_obj
        self._coords: xr.Coordinates = self._obj.coords

        if "colats" not in self._coords and "lats" in self._coords:
            self._coords = self._coords.assign(colats=90 - self._coords["lats"])

        if "mlts" not in self._coords and "longs" in self._coords:
            self._coords = self._coords.assign(
                mlts=(self._coords["longs"] + 180) * 24 / 360,
            )

    def __repr__(self) -> str:
        return "OpenGGCM accessor\n" + repr(self._coords)

    @property
    def coords(self) -> xr.Coordinates:
        return self._coords

    def __getattr__(self, name: str) -> Any:
        if name in self._coords:
            return self._coords[name]

        msg = f"Attribute {name} not found"
        raise AttributeError(msg)
